/**
 * CheapWAV.java: quick analysis of WAVE files.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-12-18 donfede@casagrau.org
 * credit to: Ringdroid, with a primary developer of Dominic Mazzoni, a Google employee and also author of Audacity. --  http://code.google.com/p/ringdroid/
 * $Id: CheapWAV.java,v 1.15 2013-03-24 19:40:21 donfede Exp $
 */

/*
 * ORIGINAL Ringdroid Apache/BSD Copyright by Google:
 * Copyright (C) 2008 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.ringdroid.soundfile;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Date;

import org.casagrau.mpq.mpq_android.R;

import android.content.SharedPreferences;
import android.os.Environment;
import android.preference.PreferenceManager;
import android.util.Log;

/**
 * CheapWAV represents a standard 16-bit WAV file, splitting it into
 * artificial frames of 20 ms and taking the maximum of each frame to
 * get an approximation of the waveform contour.
 */
public class CheapWAV extends CheapSoundFile {
	private static final String LOGTAG = "CheapWAV";
	
	public static Factory getFactory() {
        return new Factory() {
            public CheapSoundFile create() {
                return new CheapWAV();
            }
            public String[] getSupportedExtensions() {
                return new String[] { "wav" };
            }
        };
    }

    // Member variables containing frame info
    private int mNumFrames;
    private int mNumNoiseFrames;
    private int[] mFrameOffsets = new int [0];
    private int[] mFrameLens = new int [0];
    private int[] mFrameGains = new int [0];
    private int[] mFrameNoise = new int[0];
    private int mFrameBytes;
    private int mFileSize;
    private int mSampleRate;
    private int mBitRate;
    private int mChannels;

    public CheapWAV() {
    }

    public int getNumFrames() {
        return mNumFrames;
    }
    
    public int getNumNoiseFrames() {
    	return mNumNoiseFrames;
    }

    public int getSamplesPerFrame() {
        return mSampleRate / 50;
    }

    public int[] getFrameOffsets() {
        return mFrameOffsets;
    }

    public int[] getFrameLens() {
        return mFrameLens;
    }

    public int[] getFrameGains() {
        return mFrameGains;
    }
    
    public int[] getFrameNoise() {
    	return mFrameNoise;
    }
    

    public int getFrameBytes() {
    	return mFrameBytes;
    }

    public int getFileSizeBytes() {
        return mFileSize;        
    }

    public int getSampleRate() {
        return mSampleRate;
    }

    public int getChannels() {
        return mChannels;
    }
    
    public int getBitRate() {
    	return mBitRate;
    }

    public String getFiletype() {
        return "WAV";
    }

    public void ReadFile(File inputFile)
            throws java.io.FileNotFoundException,
                   java.io.IOException {
        super.ReadFile(inputFile); 
    	Log.d(LOGTAG, "ReadFile() 00  inputFile: " + inputFile);

        // Try to read meta data from a cached file.
        boolean lrc = ReadFileMeta(inputFile);
        if (lrc) return;
        
        mFileSize = (int)inputFile.length();
        if (mFileSize < 128) {
            throw new java.io.IOException("File too small to parse");
        }

        FileInputStream stream = new FileInputStream(inputFile);
        int lOffset = 0;
        byte[] header = new byte[12];
        stream.read(header, 0, 12);
        lOffset += 12;
        if (header[0] != 'R' ||
            header[1] != 'I' ||
            header[2] != 'F' ||
            header[3] != 'F' ||
            header[8] != 'W' ||
            header[9] != 'A' ||
            header[10] != 'V' ||
            header[11] != 'E') {
            throw new java.io.IOException("Not a WAV file");
        }

        mChannels = 0;
        mSampleRate = 0;
        mBitRate = 0;
        // Chunk read loop.
        while (lOffset + 8 <= mFileSize) {
            byte[] chunkHeader = new byte[8];
            stream.read(chunkHeader, 0, 8);
            lOffset += 8;

            int chunkLen =
                ((0xff & chunkHeader[7]) << 24) |
                ((0xff & chunkHeader[6]) << 16) |
                ((0xff & chunkHeader[5]) << 8) |
                ((0xff & chunkHeader[4]));
            
            // Read fmt chunk.
            if (chunkHeader[0] == 'f' &&
                chunkHeader[1] == 'm' &&
                chunkHeader[2] == 't' &&
                chunkHeader[3] == ' ') {
                if (chunkLen < 16 || chunkLen > 1024) {
                    throw new java.io.IOException(
                        "WAV file has bad fmt chunk");
                }

                byte[] fmt = new byte[chunkLen];
                stream.read(fmt, 0, chunkLen);
                lOffset += chunkLen;

                int format =
                    ((0xff & fmt[1]) << 8) |
                    ((0xff & fmt[0]));
                mChannels =
                    ((0xff & fmt[3]) << 8) |
                    ((0xff & fmt[2]));
                mSampleRate =
                    ((0xff & fmt[7]) << 24) |
                    ((0xff & fmt[6]) << 16) |
                    ((0xff & fmt[5]) << 8) |
                    ((0xff & fmt[4]));
                // 4n AvgBytesPerSec [11 -8]
                // 2n BlockAlign [13-12]
                mBitRate = 
                	((0xff & fmt[15]) << 8) |
                	((0xff & fmt[14]));

                if (format != 1) {
                    throw new java.io.IOException(
                        "Unsupported WAV file encoding");
                }

            // Read data chunk.
            } else if (chunkHeader[0] == 'd' &&
                       chunkHeader[1] == 'a' &&
                       chunkHeader[2] == 't' &&
                       chunkHeader[3] == 'a') {
                if (mChannels == 0 || mSampleRate == 0) {
                    throw new java.io.IOException(
                        "Bad WAV file: data chunk before fmt chunk");
                }

                int frameSamples = (mSampleRate * mChannels) / 50;
                mFrameBytes = frameSamples * (mBitRate / 8);
            	Log.d(LOGTAG, "ReadFile() 35  sampleR: " + mSampleRate + " chan: " + mChannels + " bitR: " + mBitRate);
            	Log.d(LOGTAG, "ReadFile() 37  framesamples: " + frameSamples + " frameBytes: " + mFrameBytes);

                mNumFrames = (chunkLen + (mFrameBytes - 1)) / mFrameBytes;
                mNumNoiseFrames = 0;
                mFrameOffsets = new int[mNumFrames];
                mFrameLens = new int[mNumFrames];
                mFrameGains = new int[mNumFrames];
                mFrameNoise = new int[mNumFrames];
                
                byte[] oneFrame = new byte[mFrameBytes];

                int i = 0;
                int frameIndex = 0;
                
                // Preparations for basic 'noise' calculations.
    			long NOISE_LIMIT_TIME = 500;
    			int NOISE_LIMIT_VOL = 100;
    			// Get audio directory from preferences.
    			if (null != mContext) {
	    			SharedPreferences sharedPref = PreferenceManager.getDefaultSharedPreferences(mContext);
	    			String lstrv = sharedPref.getString(mContext.getString(R.string.pref_noise_vol_limit_key), mContext.getString(R.string.pref_noise_vol_limit_default));
	    			String lstrt = sharedPref.getString(mContext.getString(R.string.pref_noise_time_limit_key), mContext.getString(R.string.pref_noise_time_limit_default));
	    			NOISE_LIMIT_VOL = Integer.parseInt(lstrv);
	    			NOISE_LIMIT_TIME = Long.parseLong(lstrt);
	    			} else Log.w(LOGTAG, "ReadFile() 39  NULL context, unable to get preferences, using defaults.");

    			Log.d(LOGTAG, "ReadFile() 40  vol limit: " + NOISE_LIMIT_VOL + " time limit: " + NOISE_LIMIT_TIME);

    			
    			long lcursorTimeMS = 0;
    			long lnoiseLastTimeMS = 0;
                ///////////////////////

                // Loop through and read 'artificial 20ms' frames, storing maxGain along the way.
                while (i < chunkLen) {
                    int oneFrameBytes = mFrameBytes;
                    if (i + oneFrameBytes > chunkLen) {
                        i = chunkLen - oneFrameBytes;
                    }

                    stream.read(oneFrame, 0, oneFrameBytes);

                    // Loop through 'frame', byte by byte, calculating maxGain.
                    int maxGain = 0;
                    for (int j = 1; j < oneFrameBytes; j += 4 * mChannels) {
                        int val = java.lang.Math.abs(oneFrame[j]);
                        if (val > maxGain) {
                            maxGain = val;
                        }
                    }

                    mFrameOffsets[frameIndex] = lOffset;
                    mFrameLens[frameIndex] = oneFrameBytes;
                    mFrameGains[frameIndex] = maxGain;
                    
                    // Basic 'noise' calculations.  Checking for 'noise' > $VolumeLevel, within $Time/Measures (~4-8seconds).
                    lcursorTimeMS = ((long)i * 1000) / mChannels / (mBitRate / 8) / mSampleRate;                    		
                    if (maxGain > NOISE_LIMIT_VOL) {
                    	// FIXME: Turning the noise flag on, should probably be throttled (log()?? or possibly retroactively) 
                    	mFrameNoise[frameIndex] = 1;  // FIXME: 'real noies' keep this code and use constants, or toss it.
                    	mNumNoiseFrames++;
                    	lnoiseLastTimeMS = lcursorTimeMS;
                    } else {
                		mFrameNoise[frameIndex] = 0;

                    	if (frameIndex > 1) {
	                    	// Check last frame if noise flag was on, and disable if appropriate.
	                    	if (1 <= mFrameNoise[frameIndex-1]) {
	                    		if ((lcursorTimeMS - lnoiseLastTimeMS) > NOISE_LIMIT_TIME)  {
	                    			mFrameNoise[frameIndex-1] = 0;
	                    			mNumNoiseFrames--;
		                    		// FIXME: create a for loop , checking last 8ms of frames, and disable their noise flags if less than vol limit.
	                    		} else {
		                        	mFrameNoise[frameIndex] = 2; // FIXME: 'lingering/rest noise' keep this code and use constants, or toss it.
		                        	mNumNoiseFrames++;
	                    		}
	                    	}
                    	} // END if (frameIndex > 1)
                    }
                	Log.d(LOGTAG, "ReadFile() 43  i: " + i + " lcursorTimeMS: " + lcursorTimeMS + " lastTimeMS: " + lnoiseLastTimeMS + " maxGain: " + maxGain + " noiseFrames: " + mNumNoiseFrames);
                    ///////////////////////

                    frameIndex++;
                    lOffset += oneFrameBytes;
                    i += oneFrameBytes;

                    if (mProgressListener != null) {
                        boolean keepGoing = mProgressListener.reportProgress(i * 1.0 / chunkLen);
                        if (!keepGoing) {
                            break;
                        }
                    }
                } // END while Loop through and read data chunk frames
            	Log.d(LOGTAG, "ReadFile() 49  i: " + i + " lcursorTimeMS: " + lcursorTimeMS + " numFrame: " + mNumFrames + " noise: " + mNumNoiseFrames);
            // Skip over unknown chunks.
            } else {
                stream.skip(chunkLen);
                lOffset += chunkLen;
            }
        } // END while chunk read loop.     
        stream.close();
        
        WriteFileMeta(inputFile);
    	Log.d(LOGTAG, "ReadFile() 99");
    } // END ReadFile()

    
    /**
     * Save the state of key data members to disk for later quick retrieval.
     * @param inputParentFile File of the parent audio file.
     * This method works in conjunction with {@link #ReadFileMeta()}.
     */
    public void WriteFileMeta(File inputParentFile) {
    	Log.d(LOGTAG, "WriteFileMeta() 00  P File: " + inputParentFile.toString());

    	String lparentFileName = inputParentFile.getName();
    	String[] lcomponents = lparentFileName.split("\\.");
        if (lcomponents.length < 2) {
        	Log.e(LOGTAG, "WriteFileMeta() 15 -> 99  EXIT ignoring error, insufficient components: " + lcomponents.toString());
            return;
        }

		File lFile = new File(inputParentFile.getParent() + File.separator + "meta_dir");
	    if (!lFile.isDirectory()) {
		    Log.e(LOGTAG, "WriteFileMeta() 16 DIR not directory - del");
	    	lFile.delete();
	    }
	    if (!lFile.exists() ) {
		    Log.w(LOGTAG, "WriteFileMeta() 17 DIR no exists - mkdir");
	    	lFile.mkdirs();
	    }
        
        String lmetaFileName = inputParentFile.getParent() + File.separator + "meta_dir" + File.separator + lcomponents[lcomponents.length-2] + ".meta";
    	File lMetaFile = new File(lmetaFileName);
        Log.d(LOGTAG, "WriteFileMeta() 22   M File: " + lMetaFile);

        // Credit to IBM page on java serializing --  http://www.ibm.com/developerworks/java/library/j-5things1/index.html
        try {
	        FileOutputStream fos = new FileOutputStream(lMetaFile);
	        ObjectOutputStream oos = new ObjectOutputStream(fos);
	        oos.writeChars("# MPQ APP meta data\n");
	        oos.writeChars("# " + new Date(System.currentTimeMillis())+ "\n");
	        oos.writeInt(2); // Current version of the MPQ meta file format.  // FIXME: this should probably go into a constant. 
	        oos.writeObject(lmetaFileName);
	        oos.writeInt(mFileSize);
	        oos.writeInt(mChannels);
	        oos.writeInt(mSampleRate);
	        oos.writeInt(mBitRate);
	        oos.writeInt(mNumFrames);
	        oos.writeInt(mNumNoiseFrames);
	        oos.writeObject(mFrameOffsets);
	        oos.writeObject(mFrameLens);
	        oos.writeObject(mFrameGains);
	        oos.writeObject(mFrameNoise);
	        oos.close();
        } catch (Exception e) {
			throw new IllegalStateException("Failed to create " + lMetaFile.toString() + " exception ex: " + e.getLocalizedMessage());
		}
        //Log.d(LOGTAG, "WriteFileMeta() 99");
    } // END WriteFileMeta(File inputParentFile)
    
    
    /**
     * Read the state of key data members from disk.
     * @param inputParentFile File of the parent audio file.
     * @return boolean flag, true if data was read from disk.
     * This method works in conjunction with {@link #WriteFileMeta()}.
     */
    @SuppressWarnings("deprecation")
	public boolean ReadFileMeta(File inputParentFile) {
    	Log.d(LOGTAG, "ReadFileMeta() 00  P File: " + inputParentFile.toString());

    	boolean lrc = false;	// Return code.

    	String lparentFileName = inputParentFile.getName();
    	String[] lcomponents = lparentFileName.split("\\.");
        if (lcomponents.length < 2) {
        	Log.e(LOGTAG, "ReadFileMeta() 15 -> 99  EXIT ignoring error, insufficient components: " + lcomponents.toString());
            return lrc;
        }
        String lmetaFileName = inputParentFile.getParent() + File.separator + "meta_dir" + File.separator + lcomponents[lcomponents.length-2] + ".meta";
    	File lMetaFile = new File(lmetaFileName);
    	long lMetaLen = lMetaFile.length();
        Log.d(LOGTAG, "ReadFileMeta() 22   M File: " + lMetaFile + " len: " + lMetaLen);

        // Check that meta file is valid.
        if ( (!lMetaFile.isFile()) || (0 == lMetaLen) ) {
        	Log.w(LOGTAG, "ReadFileMeta() 25 -> 99  EXIT ignoring error, invalid meta file");
        	return lrc;
        }
        
        //Log.d(LOGTAG, "ReadFileMeta() 30  original object: " + this.toString() );
        String lheader = "";
        String ldatestamp = "";
        String lreadFileName = "";
        int lmetaVersion = 0;
        // Credit to IBM page on java serializing --  http://www.ibm.com/developerworks/java/library/j-5things1/index.html
        try {
	        FileInputStream fis = new FileInputStream(lMetaFile);
	        ObjectInputStream ois = new ObjectInputStream(fis);
	        lheader = ois.readLine();
	        ldatestamp = ois.readLine();
	        lmetaVersion = ois.readInt();
	        
	        if (2 == lmetaVersion) {	        
		        lreadFileName = (String) ois.readObject();
		        mFileSize = ois.readInt();
		        mChannels = ois.readInt();
		        mSampleRate = ois.readInt();
		        mBitRate = ois.readInt();
		        mNumFrames = ois.readInt();
		        mNumNoiseFrames = ois.readInt();
                if (mProgressListener != null) {	// FIXME: this listener check sure feels like a function...
                    boolean keepGoing = mProgressListener.reportProgress(0.10);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameOffsets = (int[]) ois.readObject();
		        if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.25);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameLens = (int[]) ois.readObject();
                if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.50);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameGains = (int[]) ois.readObject();
                if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.75);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameNoise = (int[]) ois.readObject();
                if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.98);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
	        } else {
	        	Log.w(LOGTAG, "ReadFileMeta() 37 -> 99  EXIT ignoring error, invalid meta file version: " + lmetaVersion);
	        	return lrc;
	        }
	        ois.close();
	        lrc = true;
        } catch (Exception e) {
            Log.e(LOGTAG, "ReadFileMeta() Faile to read meta file: " + lMetaFile.toString() + " exception ex: " + e.getLocalizedMessage() );
            e.printStackTrace();
        }        
//        Log.d(LOGTAG, "ReadFileMeta() 41 header: " + lheader);
//        Log.d(LOGTAG, "ReadFileMeta() 43 datestamp: " + ldatestamp);
//        Log.d(LOGTAG, "ReadFileMeta() 44 version: " + lmetaVersion);
//        Log.d(LOGTAG, "ReadFileMeta() 45 read filename: " + lreadFileName);
//        Log.d(LOGTAG, "ReadFileMeta() 47  read object: " + this.toString() );

        return lrc;
    } // END WriteFileMeta(File inputParentFile)
    
    
    
    public void WriteFile(File outputFile, int startFrame, int numFrames)
            throws java.io.IOException {
    	/*
        outputFile.createNewFile();
        FileInputStream in = null; // new FileInputStream(inputFile);
        FileOutputStream out = new FileOutputStream(outputFile);

        long totalAudioLen = 0;
        for (int i = 0; i < numFrames; i++) {
            totalAudioLen += mFrameLens[startFrame + i];
        }

        long totalDataLen = totalAudioLen + 36;
        long longSampleRate = mSampleRate;
        long byteRate = mSampleRate * 2 * mChannels;

        byte[] header = new byte[44];
        header[0] = 'R';  // RIFF/WAVE header
        header[1] = 'I';
        header[2] = 'F';
        header[3] = 'F';
        header[4] = (byte) (totalDataLen & 0xff);
        header[5] = (byte) ((totalDataLen >> 8) & 0xff);
        header[6] = (byte) ((totalDataLen >> 16) & 0xff);
        header[7] = (byte) ((totalDataLen >> 24) & 0xff);
        header[8] = 'W';
        header[9] = 'A';
        header[10] = 'V';
        header[11] = 'E';
        header[12] = 'f';  // 'fmt ' chunk
        header[13] = 'm';
        header[14] = 't';
        header[15] = ' ';
        header[16] = 16;  // 4 bytes: size of 'fmt ' chunk
        header[17] = 0;
        header[18] = 0;
        header[19] = 0;
        header[20] = 1;  // format = 1
        header[21] = 0;
        header[22] = (byte) mChannels;
        header[23] = 0;
        header[24] = (byte) (longSampleRate & 0xff);
        header[25] = (byte) ((longSampleRate >> 8) & 0xff);
        header[26] = (byte) ((longSampleRate >> 16) & 0xff);
        header[27] = (byte) ((longSampleRate >> 24) & 0xff);
        header[28] = (byte) (byteRate & 0xff);
        header[29] = (byte) ((byteRate >> 8) & 0xff);
        header[30] = (byte) ((byteRate >> 16) & 0xff);
        header[31] = (byte) ((byteRate >> 24) & 0xff);
        header[32] = (byte) (2 * mChannels);  // block align
        header[33] = 0;
        header[34] = 16;  // bits per sample
        header[35] = 0;
        header[36] = 'd';
        header[37] = 'a';
        header[38] = 't';
        header[39] = 'a';
        header[40] = (byte) (totalAudioLen & 0xff);
        header[41] = (byte) ((totalAudioLen >> 8) & 0xff);
        header[42] = (byte) ((totalAudioLen >> 16) & 0xff);
        header[43] = (byte) ((totalAudioLen >> 24) & 0xff);
        out.write(header, 0, 44);

        byte[] buffer = new byte[mFrameBytes];
        int pos = 0;
        for (int i = 0; i < numFrames; i++) {
            int skip = mFrameOffsets[startFrame + i] - pos;
            int len = mFrameLens[startFrame + i];
            if (skip < 0) {
                continue;
            }
            if (skip > 0) {
                in.skip(skip);
                pos += skip;
            }
            in.read(buffer, 0, len);
            out.write(buffer, 0, len);
            pos += len;
        }

        in.close();
        out.close();
    */
    } // END WriteFile()
    
    
    /**
     * Return a string representation of this object.
     * @return String with description of object contents.
     */
    public String toString() {
        String lstr = new String();    
    	lstr = lstr + " mFileSize: " + mFileSize;
    	lstr = lstr + " mChannels: " + mChannels;
    	lstr = lstr + " mSampleRate: " + mSampleRate;
    	lstr = lstr + " mBitRate: " + mBitRate;
    	lstr = lstr + " mNumFrames: " + mNumFrames;
    	lstr = lstr + " nNumNoiseFrames: " + mNumNoiseFrames;
    	lstr = lstr + " mFrameOffsets[" + mFrameOffsets.length + "]: " + mFrameOffsets.toString();
    	lstr = lstr + " mFrameLens[" + mFrameLens.length + "]: " + mFrameLens.toString();
    	lstr = lstr + " mFrameGains[" + mFrameGains.length + "]: " + mFrameGains.toString();
    	lstr = lstr + " mFrameNoise[" + mFrameNoise.length + "]: " + mFrameNoise.toString();
    	return lstr;
    } // END toString()
    
} // END class CheapWAV

