/**
 * MPQAudioService.java: Service to control audio recording and playback.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-11-14 donfede@casagrau.org
 * credit to: android API Guide on Services -- (Apache2) http://developer.android.com/guide/components/services.html , http://developer.android.com/guide/components/bound-services.html
 *			  Reto Meier, EMEA Android Developer Advocate, Google UK; for a record and playback example -- http://eurodev.blogspot.com/2009/09/raw-audio-manipulation-in-android.html
 *			  Stjepan Rajko, urbanSTEW -- (GPL) RehearsalAssistant/RehearsalAudioRecorder.java
 *			  cristi (Cristian Baita) -- http://indyvision.net/2010/02/android-threads-tutorial-part-3/
 *			  davidkuo -- http://codinghard.wordpress.com/2009/05/16/android-thread-messaging/
 * $Id: MPQAudioService.java,v 1.20 2013-03-25 02:08:18 donfede Exp $
 */

package org.casagrau.mpq.audio_service;

import java.io.File;
import java.io.RandomAccessFile;
import java.util.HashMap;

import org.casagrau.mpq.mpq_android.MPQMainTabActivity;
import org.casagrau.mpq.mpq_android.R;
import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.annotation.SuppressLint;
import android.app.Notification;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.media.AudioFormat;
import android.media.AudioManager;
import android.os.Binder;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.IBinder;
import android.os.Message;
import android.support.v4.app.NotificationCompat;
import android.util.Log;

@SuppressLint("HandlerLeak")
public class MPQAudioService extends Service {
	private static final String LOGTAG = "MPQAudioService";
	
	/**
	 * Intent actions for Start Service and Bind Service.
	 */
	public static final String ACTION_REC = "org.casagrau.mpq.audio_service.intent.REC";				// Intent Action to begin recording a file; requires a filename to create (or overwrite).
	public static final String ACTION_PAUSE_PLAY = "org.casagrau.mpq.audio_service.intent.PAUSE_PLAY";// Intent Action to pause/un-pause playing.
	public static final String ACTION_PAUSE_REC = "org.casagrau.mpq.audio_service.intent.PAUSE_REC";	// Intent Action to pause/un-pause recording.
	public static final String ACTION_STOP = "org.casagrau.mpq.audio_service.intent.STOP";			// Intent Action to stop playing or recording.
	public static final String ACTION_PLAY = "org.casagrau.mpq.audio_service.intent.PLAY";			// Intent Action to begin playing a file; either requires a filename to play, or a previous called to open or play a file.
	public static final String ACTION_OPEN = "org.casagrau.mpq.audio_service.intent.OPEN";			// Intent Action to open a file (for playing later); requires an int with frame to seek, and expect a previous call to open or play a file.
	public static final String ACTION_SEEK = "org.casagrau.mpq.audio_service.intent.SEEK";			// Intent Action to seek to a frame within a file; requires a filename to delete.
	public static final String ACTION_AA = "org.casagrau.mpq.audio_service.intent.AA";				// Intent Action to begin audio analysis of the current open file; Expects a valid open file.
	public static final String ACTION_DEL = "org.casagrau.mpq.audio_service.intent.DEL";			// Intent Action to delete a file; requires a filename to delete.
	public static final String ACTION_BIND = "org.casagrau.mpq.audio_service.intent.BIND";			// Intent Action to bind to the service.

	// example/RecForge notifications/states: Ready, Recording, Paused_Play, Paused_Rec, Stop, Play, Init 
	public static final int FLG_SRV_STATE_INIT = -1;		// Service is initializing.
	public static final int FLG_SRV_STATE_READY = 0;		// Service is ready.
	public static final int FLG_SRV_STATE_REC = 1;			// Service is recording.
	public static final int FLG_SRV_STATE_PAUSE_PLAY = 2;	// Service is paused from playing.
	public static final int FLG_SRV_STATE_PAUSE_REC = 3;	// Service is paused from recording.
	public static final int FLG_SRV_STATE_STOP = 4;			// Service is stopping.
	public static final int FLG_SRV_STATE_PLAY = 5;			// Service is playing.
	
	public static final String ARG_FILENAME = "org.casagrau.mpq.audio_service.audio_service_filename"; 	// Argument handle for filename.
	public static final String ARG_FILEPOS  = "org.casagrau.mpq.audio_service.audio_service_fileposition";// Argument handle for cursor position in file, in milliseconds.
	
	public static final String ARG_PMSG_STATE = "state"; 		// Argument for parent messages with new state.
	public static final String ARG_PMSG_FILEPOS = "filepos"; 	// Argument for parent messages with new cursor position in milliseconds.
	public static final String ARG_PMSG_FILELEN  = "filelength";// Argument for parent messages with new *Audio* file length, in milliseconds.
	public static final String ARG_PMSG_NOISELEN  = "noiselength";// Argument for parent messages with new *Audio* noise length, in milliseconds.
	public static final String ARG_PMSG_FILENAME = "filename";	// Argument for parent messages with filename being reported on.
	public static final int ARG_PWHAT_RECPLAY = 0;				// Handler identifier for record and playback messages.
	public static final int ARG_PWHAT_AA = 1;					// Handler identifier for audio analysis messages.
	
	private static final int NOTIFICATION_ID = R.string.notification_aservice;

	
	private int mState = FLG_SRV_STATE_INIT;		// State of the audio service.  See FLG_SRV_STATE_INIT and related constants for options.
	private String mFilename;						// Filename of the file being worked on by the audio service.
	private long mFilepos;							// File position of the cursor, in milliseconds.
	private long mFilelen;							// File *Audio* length, in milliseconds.
	private int mVolumeLevel;						// Latest volume level (from mic or speaker).
	private HashMap<String, MPQAudioServiceListener> mStateListeners;	// tag/object Hash with Handles to a state change listeners in the associated activities or fragments.
	private IBinder mBinder;						// Copy of the binder interface to hand out.
	private AudioManager mAM;						// Object handle to the audio manager, used to get and release focus.
	private MPQAudioWorkerThread mWorkerThread;		// Handle to worker thread	
	private Handler mParentHandler;					// Parent message handler (this), to receive messages from child threads.
    
    
	//////////////
	// Callback listener interface
	/**
	 * Define callback interface to send messages back to the calling activity.
	 * FIXME: should this be replaced with the android message/handler system at some point (or broadcast/receiver); it is working fine for now.
	 */
	public interface MPQAudioServiceListener {
		public void onMPQAudioServiceStateChange(int newState);
		public void onMPQAudioServiceAudioChange(long cursorPos, long fileLen, int volumeLevel);
		public void onMPQAudioServiceAADone(String fileName, long fileLen, long noiseLen);
	}
	
    
	/*****************************************
     * Overridden Service methods
     */

	@Override
    public void onCreate() {
        Log.v(LOGTAG, "onCreate() 00 service created");
        mAM = (AudioManager) getSystemService(Context.AUDIO_SERVICE);
        
    	mFilename = null;
    	mFilepos = -1;
    	mFilelen = -1;
    	mVolumeLevel = -1;
        
        // Service message Handler, receiving messages from child worker threads.
        mParentHandler = new Handler() {	
        	public void handleMessage(Message msg) {
        		Log.i(LOGTAG, "mParentHndler handleMessage() msg: " + msg + " obj: " + (String)msg.obj);

        		Bundle bundle = msg.getData();
        		if (null != bundle) {
        			// Handle messages from record/worker child threads.
        			if (ARG_PWHAT_RECPLAY == msg.what) {
	        			int lstate = bundle.getInt(ARG_PMSG_STATE, -1);
	        			Log.d(LOGTAG, "mParentHandler handleMessage() 12 new lstate: " + lstate + " old mState: " + mState);
	        			if ( (lstate >= 0) && (lstate != mState) ) {
	        				mState = lstate;
	        				triggerStateListeners();
	        			}
	
	        			boolean lupdateFlag = false;
	        			long laudioPos = bundle.getLong(ARG_PMSG_FILEPOS, -1);
	        			Log.d(LOGTAG, "mParentHandler handleMessage() 14 new audioPos: " + laudioPos + " old audioPos: " + mFilepos);
	        			if ( (laudioPos >= 0) && (laudioPos != mFilepos) ) {
	        				mFilepos = laudioPos;
	        				if (FLG_SRV_STATE_REC == mState) // If recording, update the file length to the current recording position.  A bit warty to have this here... but the alternative is to make 2x calls from worker, or make a new method to send Pos + Len together.
	        					mFilelen = mFilepos;
	            			lupdateFlag = true;
	        			}
	           			long lfileLen = bundle.getLong(ARG_PMSG_FILELEN, -1);
	        			Log.d(LOGTAG, "mParentHandler handleMessage() 16 new filelen: " + lfileLen+ " old filelen: " + mFilelen);
	        			if ( (lfileLen >= 0) && (lfileLen != mFilelen) ) {
	        				mFilelen = lfileLen;
	            			lupdateFlag = true;
	        			}
	    				if (true == lupdateFlag)
	    					triggerAudioListeners();
        			} else if (ARG_PWHAT_AA == msg.what) {
            			// Handle messages from audio analysis child threads.	        			Log.i(LOGTAG, "mParentHandler handleMessage() 17 AA DONE");
        				String lfileName = bundle.getString(ARG_PMSG_FILENAME);
        				long laudioLenMS = bundle.getLong(ARG_PMSG_FILELEN);
	        			long lnoiseLenMS = bundle.getLong(ARG_PMSG_NOISELEN);
    					triggerAAListeners(lfileName, laudioLenMS, lnoiseLenMS);
        			} else Log.e(LOGTAG, "mParentHandler handleMessage() 19 unexpected what: " + msg.what);
    				
        		} else Log.e(LOGTAG, "mParentHandler handleMessage() 23 no bundle");
        	}
        }; // END mParentHandler Handler()
        
    } // END onCreate()
    
    /**
     * The service is starting, due to a call to startService().
     * Handle one of the valid option type for the Audio Service.
     * @returns int with code for START_CONTINUATION_MASK
	 * @throws IllegalArgumentException if an invalid Intent action is received, 
	 * 									or if getExtras() arguments are missing.
	 * 
     * This method aligns with, and enforces, the MPQAudioService design and state diagram.
     */
    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        Log.v(LOGTAG, "onStartCommand() 00 intent: " + intent + " flags: " + flags + " start id: " + startId);

        String lAction = intent.getAction();
        Log.d(LOGTAG, "onStartCommand() 10 intent action: " + lAction);        
        
        // Parse the Intent action, check permissions, and do the appropriate activity.
        if (lAction.equals(ACTION_REC)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_INIT != mState) &&
        			(FLG_SRV_STATE_READY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	getAudioFocus();
            showForegroundNotification(R.string.notification_aservice_rec);
			// set state to recording
        	mState = FLG_SRV_STATE_REC;
        	mFilepos = 0;
        	mFilelen = 0;
        	// Get file from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		String lFilename = null;
    		long lFilepos = 0;
    		if (null != lextraArgs) {
    			lFilename = lextraArgs.getString(MPQAudioService.ARG_FILENAME);
    			if (lFilename != null) mFilename = lFilename;
    			lFilepos = lextraArgs.getLong(MPQAudioService.ARG_FILEPOS, -1);
    			if (lFilepos >= 0) mFilepos = lFilepos;
    		} else throw new IllegalArgumentException("onStartCommand() REC missing filename.  state: " + mState + " lIAction: " + lAction);
        	// Start worker thread to record "indefinitely".
    		mWorkerThread = new MPQAudioWorkerRecThread(mParentHandler, mFilename, mFilepos, mState, (Context) this);
            mWorkerThread.start();        		
        } else if (lAction.equals(ACTION_PAUSE_REC)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_PAUSE_REC != mState) &&
        			(FLG_SRV_STATE_REC != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// Toggle of pause button, while recording.
        	if (FLG_SRV_STATE_PAUSE_REC == mState) {        		
        		getAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_rec);
                // set state to resume recording
        		mState = FLG_SRV_STATE_REC;
        	} else {
        		releaseAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_pause_rec);
            	// set state to paused
        		mState = FLG_SRV_STATE_PAUSE_REC;
        	}
        	// send new state to child worker thread 
        	sendCMsgState(mState);
        } else if (lAction.equals(ACTION_PAUSE_PLAY)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_PAUSE_PLAY != mState) &&
        			(FLG_SRV_STATE_PLAY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// Toggle of pause button, while playing.
        	if (FLG_SRV_STATE_PAUSE_PLAY == mState) {
        		getAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_play);
        		// set state to resume playing
        		mState = FLG_SRV_STATE_PLAY;
        	} else {
        		releaseAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_pause_play);
	        	// set state to paused
            	mState = FLG_SRV_STATE_PAUSE_PLAY;
        	}
        	// send new state to child worker thread 
        	sendCMsgState(mState);
        } else if (lAction.equals(ACTION_PLAY)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_INIT != mState) &&
        			(FLG_SRV_STATE_READY != mState)	) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// get file from intent "extra" (optional, provided a file has been pre-opened/played/recorded)
        	getAudioFocus();
            showForegroundNotification(R.string.notification_aservice_play);
        	// set state to playing
        	mState = FLG_SRV_STATE_PLAY;
        	// Get file from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		String lFilename = null;
    		long lFilepos = -1;
    		if (null != lextraArgs) {
    			lFilename = lextraArgs.getString(MPQAudioService.ARG_FILENAME);
    			lFilepos = lextraArgs.getLong(MPQAudioService.ARG_FILEPOS, -1);
    		} 
        	if (null == mWorkerThread) {
    			if (lFilename != null) mFilename = lFilename;
    			if (lFilepos >= 0) mFilepos = lFilepos;
            	mFilelen = -1; // Reset audio length, to be updated by the playback thread in moments.
    			// Start worker thread to play until end of file.
    			mWorkerThread = new MPQAudioWorkerPlayThread(mParentHandler, mFilename, mFilepos, mState, (Context) this);
                mWorkerThread.start();
        	} else {
            	sendCMsgState(mState);
            	if (lFilepos >= 0) sendCMsgPos(lFilepos);
        	}
        } else if (lAction.equals(ACTION_STOP)) {
        	// No valid state check; always accept stop.  Stop notification and stop foreground service
        	stopForeground(true);
        	releaseAudioFocus();
        	// set state to stop 
        	mState = FLG_SRV_STATE_STOP;
        	mFilename = null;
        	mFilepos = -1;
        	mFilelen = -1;
        	mVolumeLevel = -1;
        	sendCMsgState(mState);
        	// Release the worker thread.
    		mWorkerThread = null;
        } else if (lAction.equals(ACTION_OPEN)) {     	
        	// Valid state check        	
        	if ( (FLG_SRV_STATE_INIT != mState) &&
        			(FLG_SRV_STATE_STOP != mState) &&
        			(FLG_SRV_STATE_READY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// Get file from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		String lFilename = null;
    		if (null != lextraArgs) {
    			lFilename = lextraArgs.getString(MPQAudioService.ARG_FILENAME);
    		} else throw new IllegalArgumentException("onStartCommand() OPEN missing filename.  state: " + mState + " lIAction: " + lAction);
        	
        	// Set state to ready if there is a valid file with the Intent.
			if (lFilename != null) {
				mFilename = lFilename;
				mFilepos = 0;
	        	mFilelen = -1;
	            showForegroundNotification(R.string.notification_aservice_ready);
	        	mState = FLG_SRV_STATE_READY;
			}
        } else if (lAction.equals(ACTION_SEEK)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_PLAY != mState) &&
        			(FLG_SRV_STATE_READY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}

        	// Get file position from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		int lFilepos = 0;
    		if (null != lextraArgs) {
    			lFilepos = lextraArgs.getInt(MPQAudioService.ARG_FILEPOS);
    		} else throw new IllegalArgumentException("onStartCommand() SEEK missing pos.  state: " + mState + " lIAction: " + lAction);

        	// send new position to child worker thread 
    		Log.d(LOGTAG, "onStartCommand() 73 SEEK OLD mFilepos: " + mFilepos + " pos: " + lFilepos);        
    		if (lFilepos > 0) {
    			mFilepos = lFilepos;
            	sendCMsgPos(lFilepos);
    		}
        } else if (lAction.equals(ACTION_AA)) {
        	// Valid state check        	
        	if ( (FLG_SRV_STATE_READY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}        	
        	// Kick off thread to run analysis, which will trigger updates when done. (no state changes; this is a fire and forget action)
        	MPQAAThread laa = new MPQAAThread(mParentHandler, mFilename, mFilepos, mState, (Context) this);
			laa.run();
        } else if (lAction.equals(ACTION_DEL)) {
        	// No valid state check; always accept delete.

        	// Get audio file name from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		String lFilename = null;
    		if (null != lextraArgs) {
    			lFilename = lextraArgs.getString(MPQAudioService.ARG_FILENAME);
    		} else throw new IllegalArgumentException("onStartCommand() DELETE missing filename.  state: " + mState + " lIAction: " + lAction);
    		// Check for, and delete the audio file. 
    		File lFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + lFilename);
    		boolean lrc = false;
            Log.d(LOGTAG, "onStartCommand() 81 DELETE file: " + lFile);        
    		// Delete any previous audio file recording.
    		if (lFile.exists())
    			lrc = lFile.delete();
            Log.d(LOGTAG, "onStartCommand() 83 DELETE file RC: " + lrc);
            // Calculate the the meta file name.
        	String[] lcomponents = lFile.getName().split("\\.");
            if (lcomponents.length < 2) {
            	Log.e(LOGTAG, "onStartCommand() 84 -> 99  DELETE ignoring meta file, insufficient components: " + lcomponents.toString());
            } else {
	            String lmetaFileName = lFile.getParent() + File.separator + "meta_dir" + File.separator + lcomponents[lcomponents.length-2] + ".meta";
	        	lFile = new File(lmetaFileName);
	        	if (lFile.exists())
	    			lrc = lFile.delete();
	        	Log.d(LOGTAG, "onStartCommand() 85 DELETE Meta file RC: " + lrc);
            }
            
            lrc = stopSelfResult(startId);
        	Log.d(LOGTAG, "onStartCommand() 87 DELETE stopSelf() RC: " + lrc);
        } else {
        	// Raise exception
        	Log.d(LOGTAG, "onStartCommand() 98 RAISING EXCEPTION");
        	throw new IllegalArgumentException("onStartCommand() lIAction: " + lAction);
        }

        // Notify listener of state change.
		triggerStateListeners();
		
        return START_STICKY; // see START_CONTINUATION_MASK for other options to return, such as START_NOT_STICKY or START_REDELIVER_INTENT
    } // END onStartCommand()
    
     
    
    //////////////////////////////////
    //
    // Binder interface for this service.
    /**
     * Binder class for the MPQ Audio Service.
     */
    public class LocalBinder extends Binder {
        /**
         * Return a handle to this instance of the Audio Service service, so clients can call public methods
         * @return MPQAudioService
         */
    	public MPQAudioService getService() {
            return MPQAudioService.this;
        }
    } // END class LocalBinder

    /**
     * Method called the first time a client binds to the service.
     * Initializes local binder and other necessary objects.
     * @param intent to bind to the service
     * @return IBinder 
     */
	@Override
	public IBinder onBind(Intent intent) {
        Log.d(LOGTAG, "onBind() intent: " + intent);
    	mBinder = new LocalBinder();
    	if (null == mStateListeners) {
        	mStateListeners = new HashMap<String,MPQAudioServiceListener>();    		
    	}
        return mBinder;
	}
	
	
	

    
    @Override
    public void onDestroy() {
        Log.d(LOGTAG, "onDestroy()");
        // The service is no longer used and is being destroyed
        releaseAudioFocus(); // just in case
    }

    
    //////////////////////////////////
    //
    // public methods for the class, used by the binder
    
    
    /**
     * Return the state of the Audio Service via the binder.
     * @return int with a code identifying the audio service state.
     */
    public int getStateBnd() {
 		Log.d(LOGTAG, "getStateBnd() 00 state: " + mState);
    	return mState;
    }
    
    /**
     * Return the current file, via the binder, if the state is REC, PLAY, PAUSE, or READY.
     * @return String with current filename if a valid state, otherwise return null.
     */
    public String getFilenameBnd() {
 		Log.d(LOGTAG, "getFilenameBnd() 00 mFilename: " + mFilename);
    	String lFilename = null;
		if ( (FLG_SRV_STATE_REC == mState) ||
				(FLG_SRV_STATE_PLAY == mState) ||
				(FLG_SRV_STATE_PAUSE_REC == mState) ||
				(FLG_SRV_STATE_PAUSE_PLAY == mState) ||
				(FLG_SRV_STATE_READY == mState) ) {
			lFilename = mFilename;
		}
    	return lFilename;
    } // END getFilenameBnd()
    
    /**
     * Return the current cursor position (ms), in the audio file (excluding headers), via the binder.
     *   State must be REC, PLAY, PAUSE, or READY.
     * @return long with current audio file position in milliseconds, or -1 if there is an error.
     */
    public long getFilePosMSBnd() {
    	Log.d(LOGTAG, "getFilePosMSBnd() 00 mFilename: " + mFilename + " mFilepos: " + mFilepos);
    	long lFilepos = -1;
    	if ( (FLG_SRV_STATE_REC == mState) ||
				(FLG_SRV_STATE_PLAY == mState) ||
				(FLG_SRV_STATE_PAUSE_REC == mState) ||
				(FLG_SRV_STATE_PAUSE_PLAY == mState) ||
				(FLG_SRV_STATE_READY == mState) ) {
			lFilepos = mFilepos;
		}
    	return lFilepos;
    } // END getFileposBnd()
    
    /**
     * Return the current volume level, via the binder.
     *   State must be REC, PLAY, PAUSE.
     * @return int with the latest amplitude.
     */
    public int getAudioVolBnd() {
    	Log.d(LOGTAG, "getAudioVolBnd() 00 mFilename: " + mFilename + " mFilepos: " + mFilepos + " mVolumeLevel: " + mVolumeLevel);
    	int lVolumeLevel = -1;
    	if ( (FLG_SRV_STATE_REC == mState) ||
				(FLG_SRV_STATE_PLAY == mState) ||
				(FLG_SRV_STATE_PAUSE_REC == mState) ||
				(FLG_SRV_STATE_PAUSE_PLAY == mState) ) {
			lVolumeLevel = mVolumeLevel;
		}
    	return lVolumeLevel;
    } // END getAudioVolBnd()
    
    
    /**
     * Return the current file length (ms), via the binder.
     * @return long with current file length in milliseconds or -1 if there is an error.
     */
    public long getFileLengthMSBnd() {
    	Log.d(LOGTAG, "getFileLengthMSBnd() 00  filename:" + mFilename + " old mFilelen: " + mFilelen);
    	//  Convert bytes to milliseconds, based on channelCount,frequency, and sampleBit. // FIXME: hard coded now, read from wave header later...
    	long lfileMS = -1;
    	
    	if (null != mFilename) {
    		if (mFilelen >= 0) {
    	    	Log.d(LOGTAG, "getFileLengthMSBnd() 09  CACHED len:" + mFilelen);
    			lfileMS = mFilelen;
    		} else {
    			try {
		    		// FIXME: this block is from MPQAudioWorkerPlayThread -- likely a foundation for WAVE reader method... (though no sharing between that scope and this ... maybe static?? ) 
					// Open file to play back
		    		File lFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + mFilename);
		    	    
					RandomAccessFile lFileHandle = new RandomAccessFile(lFile, "r");
		
					// FIXME: read WAVE file headers
					int lfrequency = 11025; // cycles per second
					double lfrequencyMS = lfrequency / 1000.0; // cycles per millisecond
					int lchannelConfiguration = AudioFormat.CHANNEL_OUT_MONO;
					int laudioEncoding = AudioFormat.ENCODING_PCM_16BIT; 
					
			    	short lchannelCount;
			    	if (AudioFormat.CHANNEL_OUT_MONO == lchannelConfiguration) lchannelCount = 1;
			    	else if (AudioFormat.CHANNEL_OUT_STEREO == lchannelConfiguration) lchannelCount = 2;
			    	else throw new IllegalStateException("getFileLengthMSBnd() unsuported channel configuration: " + lchannelConfiguration);
		
			    	short lsampleBit; // sample bitrate
			    	if (AudioFormat.ENCODING_PCM_16BIT == laudioEncoding) lsampleBit = 16;
			    	else if (AudioFormat.ENCODING_PCM_8BIT == laudioEncoding) lsampleBit = 8;
			    	else throw new IllegalStateException("getFileLengthMSBnd() unsuported sample bitrate: " + lchannelConfiguration);
			    	
			    	int lframeSize = (lsampleBit/8)*lchannelCount;
			    	//Log.d(LOGTAG, "getFileLengthMSBnd() 20  framesize:" + lframeSize + " channelCount: " + lchannelCount + " sampleBit: " + lsampleBit);
			    	
			    	long lfilebytes = lFileHandle.length();
			    	//Log.d(LOGTAG, "getFileLengthMSBnd() 22  filebytes:" + lfilebytes + " fileb - head: " + (lfilebytes - 40) );
			    	lfilebytes = lfilebytes - 40; // FIXME: crappy ballpark guess at headers -- this should be replaced with actually reading wave headers and subtracting from size OR get WAVE data chunk size
			    	long lframeCount = lfilebytes / lframeSize;
			    	
			    	lfileMS = (long) (lframeCount / lfrequencyMS);
			    	//Log.d(LOGTAG, "getFileLengthMSBnd() 24  fileMS: " + lfileMS + " frameCount: " + lframeCount + " lfreq: " + lfrequency + " lfreqMS: " + lfrequencyMS);
		    	} catch (Exception e) {
		        	Log.w(LOGTAG, "getFileLengthMSBnd() 55  EXCEPTION e:" + e.getLocalizedMessage());
					e.getLocalizedMessage();
		    	}
    			mFilelen = lfileMS;
    		}
    		
    	} // END if mFilename != null
    	Log.d(LOGTAG, "getFileLengthMSBnd() 99 lfileMS: " + lfileMS);
    	return lfileMS;
    } // END getFileLengthMSBnd()

    
    /**
     * Configure the state change listener, via the binder, allowing an activity to react to state changes.
     * Expects onBind() to have been called, initializing mStateListeners; this is a binder method.
     * @param newListener with an MPQAudioServiceListener
     * @param tag with a String identifying the owner of the Listener, such as Fragment.getTag().
     * 			Currently needs to be either R.string.frag_tab2_detail or frag_audio_ui.
     */
    public void setListenerBnd(MPQAudioServiceListener newListener, String tag) {
 		Log.d(LOGTAG, "setListenerBnd() 00 newListener: " + newListener + " tag " + tag + " old size listeners: " + mStateListeners.size() );
 		if (null != tag) {
 	 		mStateListeners.put(tag, newListener); 			
 		} else {
 	 		Log.w(LOGTAG, "setListenerBnd() 10 NULL tag: " + tag + " NOT adding listener"); 			
 		}
    } // END setListenerBnd()
    	
    /**
     * Removes a state change listener, via the binder.
     * Expects onBind() to have been called, initializing mStateListeners; this is a binder method.
     * @param tag with a String identifying the owner of the Listener, such as Fragment.getTag().
     */
    public void delListenerBnd(String tag) {
 		Log.d(LOGTAG, "delListenerBnd() 00 tag: " + tag + " old size listeners: " + mStateListeners.size() );
 		if (null != tag) {
 	 		mStateListeners.remove(tag); 			
 		} else {
 	 		Log.w(LOGTAG, "delListenerBnd() 10 NULL tag: " + tag + " NOT removing listener"); 			
 		}
    } // END delListenerBnd()

    
    
    //////////////////////////////////
    //
    // private worker methods for the class, including sub-class MPQAudioThread
    
    /**
     * Show a notification and start running foreground service.
     * @param noticeID int code with an ID for the notification content to display
     * 		    ex: R.string.notification_aservice_rec , defined in strings.xml
	 * @throws IllegalArgumentException if an noticeID is received
     */
    private void showForegroundNotification(int noticeID) {
    	CharSequence ltext;
    	CharSequence ltitle = getText(R.string.notification_aservice);

    	if (R.string.notification_aservice_rec == noticeID) {
    		ltext = getText(R.string.notification_aservice_rec);
    	} else if (R.string.notification_aservice_play == noticeID) {
    		ltext = getText(R.string.notification_aservice_play);
    	} else if (R.string.notification_aservice_pause_rec == noticeID) {
    		ltext = getText(R.string.notification_aservice_pause_rec);
    	} else if (R.string.notification_aservice_pause_play == noticeID) {
    		ltext = getText(R.string.notification_aservice_pause_play);
    	} else if (R.string.notification_aservice_ready == noticeID) {
    		ltext = getText(R.string.notification_aservice_ready);
    	} else {
        	// Raise exception
        	Log.e(LOGTAG, "showNotification() 20 RAISING EXCEPTION");
        	throw new IllegalArgumentException("showNotification() noticeID: " + noticeID);
    	}
    	
        // Set the icon, scrolling text and timestamp
    	NotificationCompat.Builder builder = new NotificationCompat.Builder(this);
    	builder.setSmallIcon(R.drawable.ic_launcher);
    	builder.setContentTitle(ltitle);
    	builder.setContentText(ltext);

    	Intent intent = new Intent(this, MPQMainTabActivity.class);
        // The PendingIntent to launch our activity if the user selects this notification
        PendingIntent lcontentIntent = PendingIntent.getActivity(this, 0, intent, 0); // FIXME: keep or toss this -- PendingIntent.FLAG_CANCEL_CURRENT);
    	builder.setContentIntent(lcontentIntent);

        // Build and send the notification and begin running service in the foreground.
        Notification n = builder.build();
        startForeground(NOTIFICATION_ID, n);
    } // END showForegroundNotification()
    

    /**
     * Request audio focus.  
     */
    private void getAudioFocus() {
 		Log.d(LOGTAG, "getAudioFocus() 00  ");
 		
 		int lrc;
 		lrc = mAM.requestAudioFocus(audioFocusListener, AudioManager.STREAM_MUSIC, AudioManager.AUDIOFOCUS_GAIN);

 		if (lrc == AudioManager.AUDIOFOCUS_REQUEST_GRANTED) {
 			;// nothing for now...
 		} else {
 	 		Log.e(LOGTAG, "getFocus() 30  -- unable to get audio focus.");  // FIXME: do something better. 	ideally a user dialog - but how to trigger from the service -- need error return/state (in addition to current state change)
 		}
 		//Log.d(LOGTAG, "getAudioFocus() 99  rc:" + lrc);
    } // END getFocus()
    
    /**
     * Callback listener for audio focus change.
     */
    private AudioManager.OnAudioFocusChangeListener audioFocusListener = new AudioManager.OnAudioFocusChangeListener() {
    	public void onAudioFocusChange(int focusChange) {
    		Log.e(LOGTAG,"onAudioFocusChange() listener called focusChange:" + focusChange);
    		// FIXME: implement response for when we lose focus -- likely pause (play and/or rec)...
            if (focusChange == AudioManager.AUDIOFOCUS_LOSS_TRANSIENT) {
            	// Pause playback
            } else if (focusChange == AudioManager.AUDIOFOCUS_GAIN) {
            	// Resume playback 
            } else if (focusChange == AudioManager.AUDIOFOCUS_LOSS) {
                releaseAudioFocus(); // FIXME: this may not be required, if we trigger the stop action... 
                // Stop playback
            }
    	}
    };
    
    /**
     * Release audio focus.
     * {@link #mAM handle to AudioManager used.
     */
    private void releaseAudioFocus() {
 		Log.d(LOGTAG, "releaseAudioFocus() 00  ");
 		mAM.abandonAudioFocus(audioFocusListener);
    }

    
    /**
     * Trigger UI listeners of state change.
     * {@link #mStateListener} checked if valid, then called.
     */
    private void triggerStateListeners() {
        Log.d(LOGTAG, "triggerStateListeners() 00 state: " + mState);
        MPQAudioServiceListener lasl;	// Local handle to an audio service listener.
        
    	if (null != mStateListeners) {
    		Log.d(LOGTAG, "triggerStateListeners() 11 size listeners: " + mStateListeners.size() );
    		String ltag;
    		for (Object s: mStateListeners.keySet()) {
    			ltag = (String) s;
        		Log.d(LOGTAG, "triggerStateListeners() 13  in for loop  ltag: " + ltag);
				lasl = mStateListeners.get(ltag);
	        	lasl.onMPQAudioServiceStateChange(mState);
	        }     
    	} // END if
    } // END triggerStateListeners()
     
    
    /**
     * Trigger UI listeners of audio changes.
     * {@link #mStateListener} checked if valid, then called.
     */
    private void triggerAudioListeners() {
        Log.d(LOGTAG, "triggerAudioListeners() 00 filename: " + mFilename + " pos: " + mFilepos + " len: " + mFilelen + " vol: " + mVolumeLevel);
        Log.d(LOGTAG, "triggerAudioListeners() 10  pos : " + SessionsDB.longSessionMSecString(mFilepos) + " len: " + SessionsDB.longSessionMSecString(mFilelen) );
        MPQAudioServiceListener lasl;	// Local handle to an audio service listener.

    	if (null != mStateListeners) {
            Log.w(LOGTAG, "triggerAudioListeners() 21  size listeners: " + mStateListeners.size() );
            String ltag;
            for (Object s: mStateListeners.keySet()) {
    			ltag = (String) s;
	            Log.d(LOGTAG, "triggerAudioListeners() 23  in loop  ltag: " + ltag );
				lasl = mStateListeners.get(ltag);
	        	lasl.onMPQAudioServiceAudioChange(mFilepos, mFilelen, mVolumeLevel);
	        }
    	} // END if
    } // END triggerAudioListeners()
     
    /**
     * Trigger audio analysis listeners of updates.
     * {@link #mStateListener} checked if valid, then called.
     */
    private void triggerAAListeners(String fileName, long audioLenMS, long noiseLenMS) {
        Log.d(LOGTAG, "triggerAAListeners() 00 mFilename: " + mFilename + " vs AA fileName: " + fileName + " audioLenMS: " + audioLenMS + " noiseLenMS: " + noiseLenMS);
        MPQAudioServiceListener lasl;	// Local handle to an audio service listener.

    	if (null != mStateListeners) {
            Log.d(LOGTAG, "triggerAAListeners() 21  size listeners: " + mStateListeners.size() );
            String ltag;
            for (Object s: mStateListeners.keySet()) {
    			ltag = (String) s;
    			Log.d(LOGTAG, "triggerAAListeners() 23  in loop  ltag: " + ltag );
				lasl = mStateListeners.get(ltag);
	        	lasl.onMPQAudioServiceAADone(fileName, audioLenMS, noiseLenMS);
	        }
    	} // END if
    } // END triggerAAListeners()
     
    

	/**
	 * Prepare and send a message to the child (worker) thread with current state.
	 * Credit: Cristian Baita
	 */
	private void sendCMsgState(int state) {
		Log.d(LOGTAG, "sendCMsgState() 00 state: " + state);        		
		Message messageToChild = new Message();
		Bundle messageData = new Bundle();
//		messageToThread.what = 0;
		messageData.putInt(MPQAudioWorkerThread.ARG_CMSG_STATE, mState);
		messageToChild.setData(messageData);

		// send message to worker thread
		if (null != mWorkerThread) {
			mWorkerThread.getHandler().sendMessage(messageToChild);
		}
		return;
	} // END sendCMsgState()
    

	/**
	 * Prepare and send a message to the child (worker) thread with new cursor position in milliseconds.
	 */
	private void sendCMsgPos(long filepos) {
		Log.d(LOGTAG, "sendCMsgPos() 00 filepos: " + filepos);
		Message messageToChild = new Message();
		Bundle messageData = new Bundle();
////		messageToParent.what = 0;
		messageData.putLong(MPQAudioWorkerThread.ARG_CMSG_FILEPOS, filepos);
		messageToChild.setData(messageData);

		// send message to worker thread
		if (null != mWorkerThread) {
			mWorkerThread.getHandler().sendMessage(messageToChild);
		}
		return;
	} // END sendCMsgPos()
	
} // END class MPQAudioService
