/**
 * MPQAudioWorkerThread.java: Parent class for worker thread used by the audio service.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-12-04 donfede@casagrau.org
 * credit to: Reto Meier, EMEA Android Developer Advocate, Google UK; for a record and playback example -- http://eurodev.blogspot.com/2009/09/raw-audio-manipulation-in-android.html
 *			  Stjepan Rajko, urbanSTEW -- (GPL) RehearsalAssistant/RehearsalAudioRecorder.java
 *			  cristi (Cristian Baita) -- http://indyvision.net/2010/02/android-threads-tutorial-part-3/
 * $Id: MPQAudioWorkerThread.java,v 1.10 2013-03-24 16:07:44 donfede Exp $
 */

package org.casagrau.mpq.audio_service;

import java.io.RandomAccessFile;

import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.content.Context;
import android.content.Intent;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;


/**
 * Parent class for worker thread used by the audio service.  
 * Known direct subclasses: MPQAudioWorkerPlayThread, MPQAudioWorkerRecThread .
 */
public abstract class MPQAudioWorkerThread extends Thread {
	private static final String LOGTAG = "MPQAudioWorkerThread";

	public static final String ARG_CMSG_STATE = "state"; 	// Argument for child messages with new state.
	public static final String ARG_CMSG_FILEPOS = "filepos";// Argument for child messages with new cursor file position in milliseconds.
	
	protected static final int AUDIO_PAUSE_MS = 100;		// Amount of time in milliseconds, to wait during pause loops.  
	protected static final int AUDIO_THROTTLE_MSG = 200;	// Throttle audio messages, in milliseconds. 

	protected Handler mParentHandler;		// Parent message handler, to receive messages from child threads.  // FIXME: confirm this is correct...
	protected Handler mChildHandler;		//  // FIXME: can the same child handler be re-used/shared by several child threads???  is that even necessary? or do we expect to have 1x child thread at a time????  

	protected String mFilename;				// Filename being worked on.
	protected long mFilepos;				// File position of the cursor, in milliseconds.
	protected long mFileposSeek;			// New file position to seek to, in milliseconds.
	protected long mFilelen;				// File *Audio* length, in milliseconds.
	protected long mFilenoise;				// File noise length, in milliseconds.
	protected int mVolumeLevel;				// Latest volume level (from mic or speaker).
	protected long mAudioThrottleTimer;     // Timer for message throttle.  
	protected RandomAccessFile mFileHandle;	// File handle for audio IO
	protected int mStateThread = MPQAudioService.FLG_SRV_STATE_INIT;	// State used 'internally' by worker threads.  MPQAudioService.FLG_SRV_STATE_INIT and related constants are used.
	protected Context mServiceContext;		// Context to the parent service, to be used to send startService() intents.
	


	
	/**  
	 * Constructor for the worker thread.
	 * @param parentHandler with a Handler from the parent thread; used to send messages back.
	 * @param filename String with the filename used by the worker thread.
	 * @param filepos long with the file position for the cursor in milliseconds.
	 * @param state int with the state.  MPQAudioService.FLG_SRV_STATE_INIT and related constants are used.
	 * @param Context with a service context, to be used later to send startService() intents. 
	 */
	public MPQAudioWorkerThread(Handler parentHandler, String filename, long filepos, int state, Context serviceContext) {
		super();
		Log.d(LOGTAG, "constructor MPQAudioWorkerThread() 00 state: " + state + " old stateThread: " + mStateThread);
		Log.d(LOGTAG, "constructor MPQAudioWorkerThread() 04 filename: " + filename + " filepos: " + filepos);    		
		mParentHandler = parentHandler;
		mFilename = filename;
		mFilepos = filepos;
		mFileposSeek = -1;
		mAudioThrottleTimer = System.currentTimeMillis();
		mStateThread = state;
		mServiceContext = serviceContext;
	}

//(for now, not required, as child Handler is declared in outer class)    	
	/**
	 * Make the worker thread Handler available to parent thread.
	 * @return Handler of child worker thread 
	 */
	protected Handler getHandler() {
		return mWorkerHandler;
	}

	// mChildHandler = new Handler(){
	// FIXME: can the same child handler be re-used/shared by several child threads???  is that even necessary? or do we expect to have 1x child thread at a time????  
	private Handler mWorkerHandler = new Handler() {
		public void handleMessage(Message msg) {
			//if (msg.what == 0){
    		Log.w(LOGTAG, "mWorkerHandler handleMessage() 00 msg: " + msg + " obj: " + (String)msg.obj);        		

    		Bundle bundle = msg.getData();
    		if (null != bundle) {
    			int lstate = bundle.getInt(ARG_CMSG_STATE, -1);
    			Log.w(LOGTAG, "mWorkerHandler handleMessage() 12 lstate: " + lstate + " stateThread: " + mStateThread);
    			if ( (lstate != -1) && (lstate != mStateThread) ) {
    				mStateThread = lstate;
    			}
    			long lfilepos = bundle.getLong(ARG_CMSG_FILEPOS, -1);
    			Log.w(LOGTAG, "mWorkerHandler handleMessage() 14 lfilepos: " + lfilepos + " mFilepos: " + mFilepos);
    			if ( (lfilepos != -1) && (lfilepos != mFilepos) ) {
    				mFileposSeek = lfilepos;
    			}
    		} else Log.e(LOGTAG, "mWorkerHandler handleMessage() 19 no bundle");
  		}
	}; // END Handler mWorkerHandler    	

	
	/*************************************************
	 * Helper methods to send messages to parent thread.
	 */
	
	/**
	 * Prepare and send a message to the parent (service) thread with child thread State.
	 */
	protected void sendPMsgState() {
		Log.w(LOGTAG, "sendPMsgState() 00 stateThread: " + mStateThread);        		
		Message messageToParent = new Message();
		Bundle messageData = new Bundle();
		messageToParent.what = MPQAudioService.ARG_PWHAT_RECPLAY; 
		messageData.putInt(MPQAudioService.ARG_PMSG_STATE, mStateThread);
		messageToParent.setData(messageData);

		// send message to mainThread
		mParentHandler.sendMessage(messageToParent);
		//Log.w(LOGTAG, "sendPMsgState() 99 ");        		
		return;
	} // END sendPMsgState()

	/**
	 * Send message to the parent (service) thread with cursor Position of the child worker thread.
	 * NOTE: This message self-throttles, dropping messages unless the audio throttle has been passed.
	 * {@link #AUDIO_THROTTLE_MSG} sets throttle value.
	 * @param cursorPos long with cursor position in milliseconds.
	 * @param ignoreThrottle boolean flag to over-ride the throttle, for essential messages that must get through (ex: one time STOP/RESET at the end). 
	 */
	protected void sendPMsgPos(long cursorPos, boolean ignoreThrottle) {
		Log.d(LOGTAG, "sendPMsgPos() 00 cursorPos: " + cursorPos + " ignoreThrottle: " + ignoreThrottle);		
		long lnowTime = System.currentTimeMillis();
		//xxLog.d(LOGTAG, "sendPMsgPos() 10 now: " + lnowTime + " throttle: " + mAudioThrottleTimer);				
		// Throttle check, with trump flag.
		if ( (true == ignoreThrottle) || (lnowTime > mAudioThrottleTimer) ) {
			mAudioThrottleTimer = lnowTime + AUDIO_THROTTLE_MSG;
			//xxLog.d(LOGTAG, "sendPMsgPos() 20      updating, new throttle: " + mAudioThrottleTimer);				
			Message messageToParent = new Message();
			Bundle messageData = new Bundle();
			messageToParent.what = MPQAudioService.ARG_PWHAT_RECPLAY; 
			messageData.putLong(MPQAudioService.ARG_PMSG_FILEPOS, cursorPos);
			messageToParent.setData(messageData);

			// send message to mainThread
			mParentHandler.sendMessage(messageToParent);
		}
		return;
	} // END sendPMsgPos()

	/**
	 * Prepare and send a message to the parent (service) thread with child thread file length.
	 * @param fileLen long with file length, in milliseconds.
	 */
	protected void sendPMsgLen(long fileLen) {
		Log.w(LOGTAG, "sendPMsgLen() 00 fileLen: " + fileLen + " mFilelen: " + mFilelen);        		
// FIXME: do we want to update the member variable if it's not the same??? -- probably look closer into this...
		Message messageToParent = new Message();
		Bundle messageData = new Bundle();
		messageToParent.what = MPQAudioService.ARG_PWHAT_RECPLAY; 
		messageData.putLong(MPQAudioService.ARG_PMSG_FILELEN, fileLen);
		messageToParent.setData(messageData);

		// send message to mainThread
		mParentHandler.sendMessage(messageToParent);
		//Log.w(LOGTAG, "sendPMsgLen() 99 ");        		
		return;
	} // END sendPMsgState()

	/**
	 * Prepare and send a message to the parent (service) thread with audio analysis results.
	 * @param fileName String with the filename being analyzed.
	 * @param fileLen long with audio length, in milliseconds.
	 * @param fileNoise long with noise length, in milliseconds.
	 */
	protected void sendPMsgAA(String fileName, long fileLen, long noiseLen) {
		Log.w(LOGTAG, "sendPMsgAA() 00 filename: " + fileName + " fileLen: " + fileLen + " noiseLen: " + noiseLen);
		Message messageToParent = new Message();
		Bundle messageData = new Bundle();
		messageToParent.what = MPQAudioService.ARG_PWHAT_AA; 
		messageData.putString(MPQAudioService.ARG_PMSG_FILENAME, fileName);
		messageData.putLong(MPQAudioService.ARG_PMSG_FILELEN, fileLen);
		messageData.putLong(MPQAudioService.ARG_PMSG_NOISELEN, noiseLen);
		messageToParent.setData(messageData);

		// send message to mainThread
		mParentHandler.sendMessage(messageToParent);
		//Log.w(LOGTAG, "sendPMsgAA() 99 ");        		
		return;
	} // END sendPMsgState()
	
	
	
	/**
	 * Send an Intent to the Audio Service to STOP.
	 * Typically called as a run() thread is about to exit.
	 */
	protected void sendSvcIntentStop() {
		Intent intent = new Intent(mServiceContext, MPQAudioService.class);
		intent.setAction(MPQAudioService.ACTION_STOP);
		mServiceContext.startService(intent);
	} // END sendSvcIntentOpen()

	/**
	 * Send an Intent to the Audio Service to OPEN the current file.
	 * Typically called as a run() thread is about to exit, to re-prep the environment.
	 * {@link #mFilename} is read for the current file.
	 */
	protected void sendSvcIntentOpen() {
		Intent intent = new Intent(mServiceContext, MPQAudioService.class);
		intent.setAction(MPQAudioService.ACTION_OPEN);
		// Pass filename via "extra-bundle"  
		Bundle lextraArgs = new Bundle();
		lextraArgs.putString(MPQAudioService.ARG_FILENAME, mFilename);
		intent.putExtras(lextraArgs);
		mServiceContext.startService(intent);
	} // END sendSvcIntentOpen()
	
	/**
	 * Send an Intent to the Audio Service to AA the current file.
	 * Typically called as a record run() thread is about to exit, or manually as needed.
	 */
	protected void sendSvcIntentAA() {
		Intent intent = new Intent(mServiceContext, MPQAudioService.class);
		intent.setAction(MPQAudioService.ACTION_AA);
		mServiceContext.startService(intent);
	} // END sendSvcIntentAA()
	
	/*************************************************
	 * General helper methods.
	 */
	
	/**
	 * Convert cursor position from bytes to milliseconds.
	 * @param cursorBytePos long with cursor position in bytes (into the audio).
	 * @param frameSize int with size of each frame in bytes.
	 * @param freqHz int with frequency in cycles/second (Hz).
	 * @return long with cursor position in milliseconds.
	 */
	protected long cursorByteToMS(long cursorBytePos, int frameSize, int freqHz) {
		//xxLog.d(LOGTAG, "cursorByteToMS() 00 cursorBytePos: " + cursorBytePos + " frameSize: " + frameSize + " freqHz: " + freqHz);        		
		double lfreqMS = freqHz / 1000.0; // cycles per millisecond
		long lframeCount = cursorBytePos / frameSize;
    	long lcursorMSPos = (long) (lframeCount / lfreqMS);
    	//xxLog.d(LOGTAG, "cursorByteToMS() 24  lcursorMSPos: " + lcursorMSPos + " frameCount: " + lframeCount + " lfreqHz: " + freqHz + " lfreqMS: " + lfreqMS);		
    	Log.d(LOGTAG, "cursorByteToMS() 99  fileMS str: " + SessionsDB.longSessionMSecString(lcursorMSPos) );		
		return lcursorMSPos;
	} // END cursorByteToMS()
	
	/**
	 * Convert cursor position in milliseconds to bytes.
	 * @param cursorMSPos long with cursor position in milliseconds (into the audio).
	 * @param frameSize int with size of each frame in bytes.
	 * @param freqHz int with frequency in cycles/second (Hz).
	 * @return long with cursor position in bytes.
	 */
	protected long cursorMSToBytes(long cursorMSPos, int frameSize, int freqHz) {
		Log.d(LOGTAG, "cursorMSToBytes() 00 cursorBytePos: " + cursorMSPos + " frameSize: " + frameSize + " freqHz: " + freqHz);        		
		double lfreqMS = freqHz / 1000.0; // cycles per millisecond
		long lframeCount = cursorMSPos * (long) lfreqMS;
    	long lcursorBytePos = (lframeCount * frameSize);
    	//xxLog.d(LOGTAG, "cursorMSToBytes() 24  fileMS: " + lfileMS + " frameCount: " + lframeCount + " lfreqHz: " + freqHz + " lfreqMS: " + lfreqMS);		
    	Log.d(LOGTAG, "cursorMSToBytes() 99  cursorBytePos: " + lcursorBytePos);		
		return lcursorBytePos; // lfileBytes;
	} // END cursorMSToBytes()
	
	
} // END class MPQAudioThread




